#!/usr/bin/python

try: import simplejson as json
except ImportError: import json
import couchdb
import httplib
import urllib
import common
import unittest


HOST = "localhost:5984"
SET_NAME = "test_suite_set_view_compact"
NUM_PARTS = 4
NUM_DOCS = 40000
DDOC = {
    "_id": "_design/test",
    "language": "javascript",
    "views": {
        "mapview1": {
            "map": "function(doc) { emit(doc.integer, doc.string); }"
        }
    }
}


class TestIncludeDocs(unittest.TestCase):

    def setUp(self):
        self._params = {
            "host": HOST,
            "ddoc": DDOC,
            "nparts": NUM_PARTS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME,
            "server": couchdb.Server(url = "http://" + HOST)
            }
        # print "Creating databases"
        common.create_dbs(self._params)
        common.populate(
            self._params,
            make_doc = lambda i: {"_id": str(i), "integer": i, "string": str(i)}
            )
        common.define_set_view(self._params, range(NUM_PARTS), [])
        # print "Databases created"


    def tearDown(self):
        # print "Deleting test data"
        common.create_dbs(self._params, True)


    def test_include_docs(self):
        # print "Querying map view"
        (map_resp, map_view_result) = common.query(self._params, "mapview1", {"include_docs": "true"})
        map_etag = map_resp.getheader("ETag")

        self.assertEqual(map_view_result["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(map_view_result["rows"]), self._params["ndocs"],
                         "Query returned %d rows" % self._params["ndocs"])

        common.test_keys_sorted(map_view_result)

        for row in map_view_result["rows"]:
            self.assertTrue("doc" in row, 'row has a "doc" property')
            doc = row["doc"]
            self.assertEqual(doc["integer"], row["key"], "doc.integer same as row.key")
            self.assertEqual(doc["string"], str(row["key"]), "doc.string same as String(row.key)")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
  Provides info about a particular server.
"""

import restclient
import simplejson
import subprocess

import string
import random

class Info:
    def __init__(self):
        self.debug = False

    def _remoteShellName(self):
        tmp = ''.join(random.choice(string.ascii_letters) for i in xrange(20))
        return 'ctl-%s@127.0.0.1' % tmp

    def runCmd(self, cmd, server, port,
               user, password, opts):
        for (o, a) in opts:
            if o == '-d' or o == '--debug':
                self.debug = True

        rest = restclient.RestClient(server, port, {'debug':self.debug})
        opts = {'error_msg': 'server-info error'}

        data = rest.restCmd('GET', '/nodes/self',
                            user, password, opts)

        json = rest.getJson(data)

        for x in ['license', 'licenseValid', 'licenseValidUntil']:
            if x in json:
                del(json[x])

        if cmd == 'get-server-info':
            return json
        elif cmd == 'server-eshell':
            name = self._remoteShellName()
            p = subprocess.call(['erl','-name',name,
                '-setcookie',json['otpCookie'],'-hidden','-remsh',json['otpNode']])
        else:
            print simplejson.dumps(json, sort_keys=True, indent=2)



import sys
import string
import socket
import select
import unittest
import threading
import time
import re
import struct

from memcacheConstants import REQ_MAGIC_BYTE, RES_MAGIC_BYTE
from memcacheConstants import REQ_PKT_FMT, RES_PKT_FMT, MIN_RECV_PACKET
from memcacheConstants import SET_PKT_FMT, DEL_PKT_FMT, INCRDECR_RES_FMT

import memcacheConstants

import moxi_mock_server

# Tests moxi multitenancy.
#
# Before you run moxi_multitenancy.py, start a moxi like...
#
#   ./moxi -z url=http://127.0.0.1:4567/pools/default/buckets/default \
#          -p 0 -U 0 -vvv -t 1 -O stderr \
#          -Z default_bucket_name=,port_listen=11333, \
#             downstream_max=1,downstream_conn_max=0,downstream_protocol=binary
#
# Then...
#
#   ruby ./t/rest_mock.rb ./t/moxi_multitenancy_rest.cfg
#
# Then...
#
#   python ./t/moxi_multitenancy.py
#
# ----------------------------------

class TestMultitenancy(moxi_mock_server.ProxyClientBase):
    def __init__(self, x):
        moxi_mock_server.ProxyClientBase.__init__(self, x)

    def testListMechs(self):
        """Test list mechs"""
        self.client_connect()

        list_req = self.packReq(memcacheConstants.CMD_SASL_LIST_MECHS);
        list_res = self.packRes(memcacheConstants.CMD_SASL_LIST_MECHS,
                                val="PLAIN")

        self.client_send(list_req)
        self.client_recv(list_res)

        self.assertTrue(self.mock_quiet())

        self.client_send(list_req)
        self.client_recv(list_res)

        self.assertTrue(self.mock_quiet())

    def testNullBucket(self):
        """Test null bucket"""
        self.client_connect()

        get_err = self.packRes(memcacheConstants.CMD_GETK,
                               status=memcacheConstants.ERR_AUTH_ERROR,
                               val="Auth failure")

        get_req = self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0')
        self.client_send(get_req)
        self.client_recv(get_err)

        get_req = self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0')
        self.client_send(get_req)
        self.client_recv(get_err)

        self.assertTrue(self.mock_quiet())

    def testNullBucketSet(self):
        """Test null bucket sets"""
        self.client_connect()

        set_req = self.packReq(memcacheConstants.CMD_SET, key='simpleSet',
                               extraHeader=struct.pack(memcacheConstants.SET_PKT_FMT, 0, 0),
                               val='123')
        set_err = self.packRes(memcacheConstants.CMD_SET,
                               status=memcacheConstants.ERR_AUTH_ERROR,
                               val="Auth failure")

        self.client_send(set_req)
        self.client_recv(set_err)

        self.client_send(set_req)
        self.client_recv(set_err)

        self.assertTrue(self.mock_quiet())

    def testGoodAuth(self):
        """Test seeing a first AUTH B2B"""
        self.client_connect()

        self.exerciseGoodAuth("bucket0", "passwordGood0")

    def testBadAuth(self):
        """Test seeing a bad AUTH B2B"""
        self.client_connect()

        auth_req = self.packReq(memcacheConstants.CMD_SASL_AUTH,
                                key="PLAIN",
                                val="\0bucket0\0passwordBad0");
        auth_res = self.packRes(memcacheConstants.CMD_SASL_AUTH,
                                status=memcacheConstants.ERR_AUTH_ERROR,
                                val="Auth failure")

        self.client_send(auth_req)
        self.client_recv(auth_res)

        self.exerciseGoodAuth("bucket0", "passwordGood0")

    def testBadPassword(self):
        """Test seeing a bad AUTH B2B"""
        self.client_connect()

        auth_req = self.packReq(memcacheConstants.CMD_SASL_AUTH,
                                key="PLAIN",
                                val="\0bucket0\0passwordBad0");
        auth_res = self.packRes(memcacheConstants.CMD_SASL_AUTH,
                                status=memcacheConstants.ERR_AUTH_ERROR,
                                val="Auth failure")

        self.client_send(auth_req)
        self.client_recv(auth_res)

        self.exerciseGoodAuth("bucket1", "passwordGood1")

    def testBadEmptyUser(self):
        """Test seeing a bad AUTH B2B"""
        self.client_connect()

        auth_req = self.packReq(memcacheConstants.CMD_SASL_AUTH,
                                key="PLAIN",
                                val="\0\0somePassword");
        auth_res = self.packRes(memcacheConstants.CMD_SASL_AUTH,
                                status=memcacheConstants.ERR_AUTH_ERROR,
                                val="Auth failure")

        self.client_send(auth_req)
        self.client_recv(auth_res)

        self.exerciseGoodAuth("bucket1", "passwordGood1")

    def testBadEmptyPassword(self):
        """Test seeing a bad AUTH B2B"""
        self.client_connect()

        auth_req = self.packReq(memcacheConstants.CMD_SASL_AUTH,
                                key="PLAIN",
                                val="\0bucket0\0");
        auth_res = self.packRes(memcacheConstants.CMD_SASL_AUTH,
                                status=memcacheConstants.ERR_AUTH_ERROR,
                                val="Auth failure")

        self.client_send(auth_req)
        self.client_recv(auth_res)

        self.exerciseGoodAuth("bucket1", "passwordGood1")

    def exerciseGoodAuth(self, user, password):
        auth_req = self.packReq(memcacheConstants.CMD_SASL_AUTH,
                                key="PLAIN",
                                val="\0" + user + "\0" + password);
        auth_res = self.packRes(memcacheConstants.CMD_SASL_AUTH,
                                val="Authenticated")

        self.client_send(auth_req)
        self.client_recv(auth_res)

        get_req = self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0');
        get_res = self.packRes(memcacheConstants.CMD_GETK,
                               status=memcacheConstants.ERR_NOT_FOUND,
                               key='keyNotThere0')

        # First time, we should see a SASL plain auth.
        self.client_send(get_req)
        self.mock_recv(auth_req)
        self.mock_send(auth_res)
        self.mock_recv(get_req)
        self.mock_send(get_res)
        self.client_recv(get_res)

        # Next time, we should not see any auth.
        self.client_send(get_req)
        self.mock_recv(get_req)
        self.mock_send(get_res)
        self.client_recv(get_res)

if __name__ == '__main__':
    unittest.main()




#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import traceback
import copy

import listservers
import buckets
import info
import util_cli as util
import mc_bin_client

import stats_buffer

class StatsCollector:
    def __init__(self, log):
        self.log = log

    def seg(self, k, v):
        # Parse ('some_stat_x_y', 'v') into (('some_stat', x, y), v)
        ka = k.split('_')
        k = '_'.join(ka[0:-1])
        kstart, kend = [int(x) for x in ka[-1].split(',')]
        return ((k, kstart, kend), int(v))

    def retrieve_node_stats(self, nodeInfo, nodeStats):
        nodeStats['portDirect'] = nodeInfo['ports']['direct']
        nodeStats['portProxy'] = nodeInfo['ports']['proxy']
        nodeStats['clusterMembership'] = nodeInfo['clusterMembership']
        nodeStats['os'] = nodeInfo['os']
        nodeStats['uptime'] = nodeInfo['uptime']
        nodeStats['version'] = nodeInfo['version']

        #memory
        nodeStats['memory'] = {}
        nodeStats['memory']['allocated'] = nodeInfo['mcdMemoryAllocated']
        nodeStats['memory']['reserved'] = nodeInfo['mcdMemoryReserved']
        nodeStats['memory']['free'] = nodeInfo['memoryFree']
        nodeStats['memory']['quota'] = nodeInfo['memoryQuota']
        nodeStats['memory']['total'] = nodeInfo['memoryTotal']

        #storageInfo
        nodeStats['StorageInfo'] = {}
        if nodeInfo['storageTotals'] is not None:

            #print nodeInfo
            hdd = nodeInfo['storageTotals']['hdd']
            if hdd is not None:
                nodeStats['StorageInfo']['hdd'] = {}
                nodeStats['StorageInfo']['hdd']['free'] = hdd['free']
                nodeStats['StorageInfo']['hdd']['quotaTotal'] = hdd['quotaTotal']
                nodeStats['StorageInfo']['hdd']['total'] = hdd['total']
                nodeStats['StorageInfo']['hdd']['used'] = hdd['used']
                nodeStats['StorageInfo']['hdd']['usedByData'] = hdd['usedByData']
            ram = nodeInfo['storageTotals']['ram']
            if ram is not None:
                nodeStats['StorageInfo']['ram'] = {}
                nodeStats['StorageInfo']['ram']['quotaTotal'] = ram['quotaTotal']
                nodeStats['StorageInfo']['ram']['total'] = ram['total']
                nodeStats['StorageInfo']['ram']['used'] = ram['used']
                nodeStats['StorageInfo']['ram']['usedByData'] = ram['usedByData']
                if ram.has_key('quotaUsed'):
                    nodeStats['StorageInfo']['ram']['quotaUsed'] = ram['quotaUsed']
                else:
                    nodeStats['StorageInfo']['ram']['quotaUsed'] = 0

        #system stats
        nodeStats['systemStats'] = {}
        nodeStats['systemStats']['cpu_utilization_rate'] = nodeInfo['systemStats']['cpu_utilization_rate']
        nodeStats['systemStats']['swap_total'] = nodeInfo['systemStats']['swap_total']
        nodeStats['systemStats']['swap_used'] = nodeInfo['systemStats']['swap_used']

        curr_items = 0
        curr_items_tot = 0
        vb_rep_curr_items = 0
        if nodeInfo['interestingStats'] is not None:
            if nodeInfo['interestingStats'].has_key('curr_items'):
                curr_items = nodeInfo['interestingStats']['curr_items']
            else:
                curr_items = 0
            if nodeInfo['interestingStats'].has_key('curr_items_tot'):
                curr_items_tot = nodeInfo['interestingStats']['curr_items_tot']
            else:
                curr_items_tot = 0
            if nodeInfo['interestingStats'].has_key('vb_replica_curr_items'):
                vb_rep_curr_items = nodeInfo['interestingStats']['vb_replica_curr_items']
            else:
                vb_rep_curr_items = 0

        nodeStats['systemStats']['currentItems'] = curr_items
        nodeStats['systemStats']['currentItemsTotal'] = curr_items_tot
        nodeStats['systemStats']['replicaCurrentItems'] = vb_rep_curr_items

    def get_hostlist(self, server, port, user, password, opts):
        try:
            opts.append(("-o", "return"))
            nodes = listservers.ListServers().runCmd('host-list', server, port, user, password, opts)

            for node in nodes:
                (node_server, node_port) = util.hostport(node['hostname'])
                node_stats = {"host" : node_server,
                          "port" : node_port,
                          "status" : node['status'],
                          "master" : server}
                stats_buffer.nodes[node['hostname']] = node_stats
                if node['status'] == 'healthy':
                    node_info = info.Info().runCmd('get-server-info', node_server, node_port, user, password, opts)
                    self.retrieve_node_stats(node_info, node_stats)
                else:
                    self.log.error("Unhealthy node: %s:%s" %(node_server, node['status']))
            return nodes
        except Exception, err:
            traceback.print_exc()
            sys.exit(1)

    def get_bucketlist(self, server, port, user, password, opts):
        try:
            bucketlist = buckets.Buckets().runCmd('bucket-get', server, port, user, password, opts)
            for bucket in bucketlist:
                bucket_name = bucket['name']
                self.log.info("bucket: %s" % bucket_name)
                bucketinfo = {}
                bucketinfo['name'] = bucket_name
                bucketinfo['bucketType'] = bucket['bucketType']
                bucketinfo['authType'] = bucket['authType']
                bucketinfo['saslPassword'] = bucket['saslPassword']
                bucketinfo['numReplica'] = bucket['replicaNumber']
                bucketinfo['ramQuota'] = bucket['quota']['ram']
                bucketinfo['master'] = server

                bucketStats = bucket['basicStats']
                bucketinfo['bucketStats'] = {}
                bucketinfo['bucketStats']['diskUsed'] = bucketStats['diskUsed']
                bucketinfo['bucketStats']['memUsed'] = bucketStats['memUsed']
                bucketinfo['bucketStats']['diskFetches'] = bucketStats['diskFetches']
                bucketinfo['bucketStats']['quotaPercentUsed'] = bucketStats['quotaPercentUsed']
                bucketinfo['bucketStats']['opsPerSec'] = bucketStats['opsPerSec']
                bucketinfo['bucketStats']['itemCount'] = bucketStats['itemCount']

                stats_buffer.bucket_info[bucket_name] = bucketinfo

                # get bucket related stats
                c = buckets.BucketStats(bucket_name)
                json = c.runCmd('bucket-stats', server, port, user, password, opts)
                stats_buffer.buckets_summary[bucket_name] = json
            return bucketlist
        except Exception, err:
            traceback.print_exc()
            sys.exit(1)

    def get_mc_stats_per_node(self, mc, stats):
        cmd_list = ["timings", "tap", "checkpoint", "memory", ""]
        #cmd_list = ["tap"]
        try:
            for cmd in cmd_list:
                node_stats = mc.stats(cmd)
                if node_stats:
                    if cmd == "timings":
                        # need to preprocess histogram data first
                        vals = sorted([self.seg(*kv) for kv in node_stats.items()])
                        dd = {}
                        totals = {}
                        longest = 0
                        for s in vals:
                            avg = (s[0][1] + s[0][2]) / 2
                            k = s[0][0]
                            l = dd.get(k, [])
                            l.append((avg, s[1]))
                            dd[k] = l
                            totals[k] = totals.get(k, 0) + s[1]
                        for k in sorted(dd):
                            ccount = 0
                            for lbl,v in dd[k]:
                                ccount += v * lbl
                            stats[k] = ccount / totals[k]
                    else:
                        for key, val in node_stats.items():
                            stats[key] = val
        except Exception, err:
            traceback.print_exc()

    def get_mc_stats(self, server, bucketlist, nodes):
        #print util.pretty_print(bucketlist)
        for bucket in bucketlist:
            bucket_name = bucket['name']
            stats_buffer.node_stats[bucket_name] = {}
            for node in nodes:
                (node_server, node_port) = util.hostport(node['hostname'])
                self.log.info("  node: %s %s" % (node_server, node['ports']['direct']))
                stats = {}
                mc = mc_bin_client.MemcachedClient(node_server, node['ports']['direct'])
                if bucket["name"] != "Default":
                    mc.sasl_auth_plain(bucket_name.encode("utf8"), bucket["saslPassword"].encode("utf8"))
                self.get_mc_stats_per_node(mc, stats)
                stats_buffer.node_stats[bucket_name][node['hostname']] = stats

    def get_ns_stats(self, bucketlist, server, port, user, password, opts):
        for bucket in bucketlist:
            bucket_name = bucket['name']
            stats_buffer.buckets[bucket_name] = copy.deepcopy(stats_buffer.stats)
            cmd = 'bucket-node-stats'
            for scale, stat_set in stats_buffer.buckets[bucket_name].iteritems():
                for stat in stat_set.iterkeys():
                    sys.stderr.write('.')
                    self.log.debug("retrieve: %s" % stat)
                    c = buckets.BucketNodeStats(bucket_name, stat, scale)

                    json = c.runCmd('bucket-node-stats', server, port, user, password, opts)
                    stats_buffer.buckets[bucket_name][scale][stat] = json
            sys.stderr.write('\n')

    def collect_data(self,cluster, user, password, opts):
        server, port = util.hostport(cluster)

        #get node list info
        nodes = self.get_hostlist(server, port, user, password, opts)
        self.log.debug(util.pretty_print(stats_buffer.nodes))

        #get bucket list
        bucketlist = self.get_bucketlist(server, port, user, password, opts)
        self.log.debug(util.pretty_print(stats_buffer.bucket_info))

        #get stats from ep-engine
        self.get_mc_stats(server, bucketlist, nodes)
        self.log.debug(util.pretty_print(stats_buffer.node_stats))

        #get stats from ns-server
        self.get_ns_stats(bucketlist, server, port, user, password, opts)
        self.log.debug(util.pretty_print(stats_buffer.buckets))




#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# source *.json files from a zip file into destination

import os
import json
import zipfile
import tempfile
import shutil

from migrator_dir import DirReader, DirWriter

sources = [{'type':'zip', 'class':'ZipReader', 'example':'zip://<zipfile>'}]
destinations = [{'type':'zip', 'class':'ZipWriter',
                 'example':'zip://<zipfile>'}]


class ZipReader(DirReader):
    def __init__(self, source):
        if source[0:2] == "//":
            source = source[2:]
        self.zipfile = zipfile.ZipFile(os.path.expanduser(source), "r")
        self.tempdir = tempfile.mkdtemp()

        self.zipfile.extractall(self.tempdir)

        DirReader.__init__(self, self.tempdir)

    def close(self):
        shutil.rmtree(self.tempdir)
        self.zipfile.close()


class ZipWriter(DirWriter):
    def __init__(self, destination):
        if destination[0:2] == "//":
            destination = destination[2:]
        self.zipfile = zipfile.ZipFile(os.path.expanduser(destination), "w",
                                       zipfile.ZIP_DEFLATED)
        self.dirname = os.path.splitext(os.path.basename(destination))[0]
        self.tempdir = tempfile.mkdtemp()

        DirWriter.__init__(self, self.tempdir)

    def close(self):
        for root, dirs, files in os.walk(self.tempdir):
            f_relative_path = os.path.abspath(root)[len(self.tempdir) + 1:]
            for f in files:
                f_path = os.path.join(root, f)
                archive_path = os.path.join(self.dirname, f_relative_path, f)
                self.zipfile.write(f_path, archive_path)

        shutil.rmtree(self.tempdir)
        self.zipfile.close()



#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

try:
    import unittest2 as unittest
except ImportError:
    import unittest
import types
import warnings

from warnings_catcher import setup_warning_catcher
from testconfig import config
from nose.plugins.attrib import attr
from couchbase.client import *
from couchbase.couchbaseclient import *


class ClientTest(unittest.TestCase):
    def setUp(self):
        self.host = config['node-1']['host']
        self.port = config['node-1']['port']
        self.username = config['node-1']['username']
        self.password = config['node-1']['password']
        self.bucket_name = config['node-1']['bucket']

    def tearDown(self):
        pass

    def setup_cb(self):
        self.cb = Couchbase(self.host + ':' + self.port,
                            self.username, self.password)

    @attr(cbv="1.0.0")
    def test_couchbase_object_construction(self):
        cb = Couchbase(self.host + ':' + self.port, self.username,
                       self.password)
        self.assertIsInstance(cb.servers, types.ListType)

    @attr(cbv="1.0.0")
    def test_server_object_construction(self):
        w = setup_warning_catcher()
        warnings.simplefilter("always")
        cb = Server(self.host + ':' + self.port, self.username, self.password)
        self.assertIsInstance(cb.servers, types.ListType)
        self.assertTrue(len(w) == 1)
        self.assertTrue("deprecated" in str(w[-1].message))

    @attr(cbv="1.0.0")
    def test_couchbase_object_construction_without_port(self):
        cb = Couchbase(self.host, self.username, self.password)
        self.assertIsInstance(cb.servers, types.ListType)

    @attr(cbv="1.0.0")
    def test_vbucketawarecouchbaseclient_object_construction(self):
        w = setup_warning_catcher()
        warnings.simplefilter("always")
        cb = VBucketAwareCouchbaseClient("http://" + self.host + ':'
                                         + self.port + "/pools/default",
                                         self.bucket_name, "")
        self.assertIsInstance(cb.servers, types.ListType)
        self.assertTrue(len(w) == 1)
        self.assertTrue("deprecated" in str(w[-1].message))

    @attr(cbv="1.0.0")
    def test_bucket(self):
        self.setup_cb()
        self.assertIsInstance(self.cb.bucket(self.bucket_name), Bucket)

    @attr(cbv="1.0.0")
    def test_buckets(self):
        self.setup_cb()
        buckets = self.cb.buckets()
        self.assertIsInstance(buckets, types.ListType)
        self.assertIsInstance(buckets[0], Bucket)

if __name__ == "__main__":
    unittest.main()



#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# source *.json files in a directory into destination

# source:
# recurses through subdirectories and reads .json files. If the .json has an
# _id field it will use that, otherwise it will use the filename. if the
# directory contains an _id file, the directory itself will be considered a
# document and all files (except *.json) will be considered json data

# destination:
# writes out items that have _id: _design/* to <dir>/design_docs
# writes out all other items to <dir>/docs

sources = [{'type':'dir', 'class':'DirReader', 'example':'dir://<directory>'}]
destinations = [{'type':'dir', 'class':'DirWriter',
                 'example':'dir://<directory>'}]

import os
import json
import migrator


class DirReader(migrator.Reader):
    def __init__(self, source):
        if source[0:2] == "//":
            source = source[2:]
        self.dir = os.path.expanduser(source)
        self.files = self._get_filenames()

    def __iter__(self):
        return self

    def _get_filenames(self):
        filenames = []
        for root, subFolders, files in os.walk(self.dir):
            for filename in files:
                if filename.endswith(".json") or filename == "_id":
                    filepath = os.path.join(root, filename)
                    filenames.append(filepath)
        return filenames

    def _get_nonjson_filenames(self, base):
        filenames = []
        for root, subFolders, files in os.walk(base):
            for filename in files:
                if not filename.endswith(".json") and not filename == "_id":
                    filepath = os.path.join(root, filename)
                    filenames.append(filepath)
        return filenames

    def next(self):
        try:
            filename = self.files.pop(0)
        except IndexError:
            raise StopIteration()

        if os.path.basename(filename) == "_id":
            # entire directory (except *.json) is a document
            # <dir>/views is converted to <id>/_views
            doc_basepath = os.path.dirname(filename)
            with open(filename, 'r') as f:
                id = f.read().strip('\n\r,')
            json_data = {}
            for f in self._get_nonjson_filenames(os.path.dirname(filename)):
                f_relative_path = f[len(doc_basepath) + 1:]
                if os.path.isfile(os.path.join(doc_basepath,
                                  f_relative_path.split(os.path.sep)[0])):
                    # its a plain file
                    with open(f) as item:
                        json_data[f_relative_path.split(os.path.sep)[0]] =\
                            item.read().strip('\n\r,')
                else:
                    # its a subdirectory
                    json_cur = json_data
                    for index in f_relative_path.split(os.path.sep)[:-1]:
                        if not index in json_cur:
                            json_cur[index] = {}
                        json_cur = json_cur[index]

                    index = os.path.splitext(f_relative_path
                                             .split(os.path.sep)[-1])[0]
                    with open(f) as item:
                        json_cur[index] = item.read().strip('\n\r,')
        else:
            with open(filename, 'r') as f:
                json_data = json.loads(f.read().strip('\n\r,'))
                if "_id" in json_data:
                    id = json_data["_id"]
                else:
                    id = os.path.splitext(os.path.basename(filename))[0]

        record = {'id': id}
        record['value'] = (dict((k, v) for (k, v) in json_data.iteritems()
                           if not k.startswith('_')))
        return record


class DirWriter(migrator.Writer):
    def __init__(self, destination):
        if destination[0:2] == "//":
            destination = destination[2:]
        self.dir = os.path.expanduser(destination)
        try:
            os.makedirs(os.path.join(destination, "docs"))
            os.makedirs(os.path.join(destination, "design_docs"))
        except OSError:
            pass

    def write(self, record):
        if os.path.sep in record["id"]:
            record["value"]["_id"] = record["id"]
        if record["id"].startswith("_design/"):
            filename = record["id"][len("_design/"):].replace(os.path.sep, "_")
            path = os.path.join(self.dir, "design_docs", filename) + ".json"
        else:
            filename = record["id"].replace(os.path.sep, "_")
            path = os.path.join(self.dir, "docs", filename) + ".json"
        f = open(path, "wb")
        f.write(json.dumps(record["value"]) + '\n')



import stats_buffer
import util_cli as util

class BucketSummary:
    def run(self, accessor):
        return  stats_buffer.bucket_info

class DGMRatio:
    def run(self, accessor):
        result = []
        hdd_total = 0
        ram_total = 0
        for node, nodeinfo in stats_buffer.nodes.iteritems():
            if nodeinfo["StorageInfo"].has_key("hdd"):
                hdd_total += nodeinfo['StorageInfo']['hdd']['usedByData']
            if nodeinfo["StorageInfo"].has_key("ram"):
                ram_total += nodeinfo['StorageInfo']['ram']['usedByData']
        if ram_total > 0:
            ratio = hdd_total / ram_total
        else:
            ratio = 0
        return ratio

class ARRatio:
    def run(self, accessor):
        result = {}
        cluster = 0
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            item_avg = {
                "curr_items": [],
                "vb_replica_curr_items": [],
            }
            num_error = []
            for counter in accessor["counter"]:
                values = stats_info[accessor["scale"]][counter]
                nodeStats = values["nodeStats"]
                samplesCount = values["samplesCount"]
                for node, vals in nodeStats.iteritems():
                    avg = sum(vals) / samplesCount
                    item_avg[counter].append((node, avg))
            res = []
            active_total = replica_total = 0
            for active, replica in zip(item_avg['curr_items'], item_avg['vb_replica_curr_items']):
                if replica[1] == 0:
                    res.append((active[0], "No replica"))
                else:
                    ratio = 1.0 * active[1] / replica[1]
                    res.append((active[0], util.pretty_float(ratio)))
                    if ratio < accessor["threshold"]:
                        num_error.append({"node":active[0], "value": ratio})
                active_total += active[1]
                replica_total += replica[1]
            if replica_total == 0:
                res.append(("total", "no replica"))
            else:
                ratio = active_total * 1.0 / replica_total
                cluster += ratio
                res.append(("total", util.pretty_float(ratio)))
                if ratio != accessor["threshold"]:
                    num_error.append({"node":"total", "value": ratio})
            #if len(num_error) > 0:
            #    result[bucket] = {"error" : num_error}
            #else:
            result[bucket] = res
        if len(stats_buffer.buckets) > 0:
            result["cluster"] = util.pretty_float(cluster / len(stats_buffer.buckets))

        return result

class OpsRatio:
    def run(self, accessor):
        result = {}
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            ops_avg = {
                "cmd_get": [],
                "cmd_set": [],
                "delete_hits" : [],
            }
            for counter in accessor["counter"]:
                values = stats_info[accessor["scale"]][counter]
                nodeStats = values["nodeStats"]
                samplesCount = values["samplesCount"]
                for node, vals in nodeStats.iteritems():
                    avg = sum(vals) / samplesCount
                    ops_avg[counter].append((node, avg))
            res = []
            read_total = write_total = del_total = 0
            for read, write, delete in zip(ops_avg['cmd_get'], ops_avg['cmd_set'], ops_avg['delete_hits']):
                count = read[1] + write[1] + delete[1]
                if count == 0:
                    res.append((read[0], "0:0:0"))
                else:
                    read_ratio = read[1] *100 / count
                    read_total += read_ratio
                    write_ratio = write[1] * 100 / count
                    write_total += write_ratio
                    del_ratio = delete[1] * 100 / count
                    del_total += del_ratio
                    res.append((read[0], "{0}:{1}:{2}".format(int(read_ratio+.5), int(write_ratio+.5), int(del_ratio+.5))))
            read_total /= len(ops_avg['cmd_get'])
            write_total /= len(ops_avg['cmd_set'])
            del_total /= len(ops_avg['delete_hits'])
            res.append(("total", "{0}:{1}:{2}".format(int(read_total+.5), int(write_total+.5), int(del_total+.5))))
            result[bucket] = res

        return result

class CacheMissRatio:
    def run(self, accessor):
        result = {}
        cluster = 0
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            values = stats_info[accessor["scale"]][accessor["counter"]]
            timestamps = values["timestamp"]
            timestamps = [x - timestamps[0] for x in timestamps]
            nodeStats = values["nodeStats"]
            samplesCount = values["samplesCount"]
            trend = []
            total = 0
            data = []
            for node, vals in nodeStats.iteritems():
                #a, b = util.linreg(timestamps, vals)
                value = sum(vals) / samplesCount
                #value = a * timestamps[-1] + b
                total += value
                trend.append((node, util.pretty_float(value)))
                data.append(value)
            total /= len(nodeStats)
            trend.append(("total", util.pretty_float(total)))
            trend.append(("variance", util.two_pass_variance(data)))
            cluster += total
            result[bucket] = trend
        if len(stats_buffer.buckets) > 0:
            result["cluster"] = util.pretty_float(cluster / len(stats_buffer.buckets))
        return result

class MemUsed:
    def run(self, accessor):
        result = {}
        cluster = 0
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            values = stats_info[accessor["scale"]][accessor["counter"]]
            timestamps = values["timestamp"]
            timestamps = [x - timestamps[0] for x in timestamps]
            nodeStats = values["nodeStats"]
            samplesCount = values["samplesCount"]
            trend = []
            total = 0
            data = []
            for node, vals in nodeStats.iteritems():
                avg = sum(vals) / samplesCount
                trend.append((node, util.size_label(avg)))
                data.append(avg)
            #print data
            trend.append(("variance", util.two_pass_variance(data)))
            result[bucket] = trend
        return result

class ItemGrowth:
    def run(self, accessor):
        result = {}
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            trend = []
            values = stats_info[accessor["scale"]][accessor["counter"]]
            timestamps = values["timestamp"]
            timestamps = [x - timestamps[0] for x in timestamps]
            nodeStats = values["nodeStats"]
            samplesCount = values["samplesCount"]
            for node, vals in nodeStats.iteritems():
                a, b = util.linreg(timestamps, vals)
                if b < 1:
                    trend.append((node, 0))
                else:
                    start_val = b
                    end_val = a * timestamps[-1] + b
                    rate = (end_val * 1.0 / b - 1.0) * 100
                    trend.append((node, util.pretty_float(rate)))
            result[bucket] = trend
        return result

class NumVbuckt:
    def run(self, accessor):
        result = {}
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            num_error = []
            values = stats_info[accessor["scale"]][accessor["counter"]]
            nodeStats = values["nodeStats"]
            for node, vals in nodeStats.iteritems():
                if vals[-1] < accessor["threshold"]:
                    num_error.append({"node":node, "value":vals[-1]})
            if len(num_error) > 0:
                result[bucket] = {"error" : num_error}
        return result

class RebalanceStuck:
    def run(self, accessor):
        result = {}
        for bucket, bucket_stats in stats_buffer.node_stats.iteritems():
            num_error = []
            for node, stats_info in bucket_stats.iteritems():
                for key, value in stats_info.iteritems():
                    if key.find(accessor["counter"]) >= 0:
                        if accessor.has_key("threshold"):
                            if int(value) > accessor["threshold"]:
                                num_error.append({"node":node, "value": (key, value)})
                        else:
                            num_error.append({"node":node, "value": (key, value)})
            if len(num_error) > 0:
                result[bucket] = {"error" : num_error}
        return result

class MemoryFramentation:
    def run(self, accessor):
        result = {}
        for bucket, bucket_stats in stats_buffer.node_stats.iteritems():
            num_error = []
            for node, stats_info in bucket_stats.iteritems():
                for key, value in stats_info.iteritems():
                    if key.find(accessor["counter"]) >= 0:
                        if accessor.has_key("threshold"):
                            if int(value) > accessor["threshold"]:
                                if accessor.has_key("unit"):
                                    if accessor["unit"] == "time":
                                        num_error.append({"node":node, "value": (key, util.time_label(value))})
                                    elif accessor["unit"] == "size":
                                        num_error.append({"node":node, "value": (key, util.size_label(value))})
                                    else:
                                        num_error.append({"node":node, "value": (key, value)})
                                else:
                                    num_error.append({"node":node, "value": (key, value)})
            if len(num_error) > 0:
                result[bucket] = {"error" : num_error}
        return result

class EPEnginePerformance:
    def run(self, accessor):
        result = {}
        for bucket, bucket_stats in stats_buffer.node_stats.iteritems():
            num_error = []
            for node, stats_info in bucket_stats.iteritems():
                for key, value in stats_info.iteritems():
                    if key.find(accessor["counter"]) >= 0:
                        if accessor.has_key("threshold"):
                            if accessor["counter"] == "flusherState" and value != accessor["threshold"]:
                                num_error.append({"node":node, "value": (key, value)})
                            elif accessor["counter"] == "flusherCompleted" and value == accessor["threshold"]:
                                num_error.append({"node":node, "value": (key, value)})
                            else:
                                if value > accessor["threshold"]:
                                    num_error.append({"node":node, "value": (key, value)})
            if len(num_error) > 0:
                result[bucket] = {"error" : num_error}
        return result

class TotalDataSize:
    def run(self, accessor):
        result = []
        total = 0
        for node, nodeinfo in stats_buffer.nodes.iteritems():
            if nodeinfo["StorageInfo"].has_key("hdd"):
                total += nodeinfo['StorageInfo']['hdd']['usedByData']
        result.append(util.size_label(total))
        return result

class AvailableDiskSpace:
    def run(self, accessor):
        result = []
        total = 0
        for node, nodeinfo in stats_buffer.nodes.iteritems():
            if nodeinfo["StorageInfo"].has_key("hdd"):
                total += nodeinfo['StorageInfo']['hdd']['free']
        result.append(util.size_label(total))
        return result

ClusterCapsule = [
    {"name" : "TotalDataSize",
     "ingredients" : [
        {
            "name" : "totalDataSize",
            "description" : "Total Data Size across cluster",
            "code" : "TotalDataSize",
        }
     ],
     "clusterwise" : True,
     "perNode" : False,
     "perBucket" : False,
    },
    {"name" : "AvailableDiskSpace",
     "ingredients" : [
        {
            "name" : "availableDiskSpace",
            "description" : "Available disk space",
            "code" : "AvailableDiskSpace",
        }
     ],
     "clusterwise" : True,
     "perNode" : False,
     "perBucket" : False,
    },
   {"name" : "CacheMissRatio",
     "ingredients" : [
        {
            "name" : "cacheMissRatio",
            "description" : "Cache miss ratio",
            "counter" : "ep_cache_miss_rate",
            "scale" : "hour",
            "code" : "CacheMissRatio",
            "unit" : "percentage",
            "threshold" : 2,
        },
     ],
     "clusterwise" : True,
     "perNode" : True,
     "perBucket" : True,
     "indicator" : False,
     "nodeDisparate" : True,
    },
    {"name" : "DGM",
     "ingredients" : [
        {
            "name" : "dgm",
            "description" : "Disk to Memory Ratio",
            "code" : "DGMRatio"
        },
     ],
     "clusterwise" : True,
     "perNode" : False,
     "perBucket" : False,
    },
    {"name" : "BucketSummary",
     "ingredients" : [
        {
            "name" : "bucketSummary",
            "description" : "Bucket performance summary",
            "code" : "BucketSummary",
        },
     ],
     "clusterwise" : True,
    },
    {"name" : "ActiveReplicaResidentRatio",
     "ingredients" : [
        {
            "name" : "activeReplicaResidencyRatio",
            "description" : "Active and Replica Resident Ratio",
            "counter" : ["curr_items", "vb_replica_curr_items"],
            "scale" : "minute",
            "code" : "ARRatio",
            "threshold" : 1,
        },
     ],
     "clusterwise" : True,
     "perNode" : True,
     "perBucket" : True,
     "indicator" : True,
    },
    {"name" : "OPSPerformance",
     "ingredients" : [
        {
            "name" : "opsPerformance",
            "description" : "Read/Write/Delete ops ratio",
            "scale" : "minute",
            "counter" : ["cmd_get", "cmd_set", "delete_hits"],
            "code" : "OpsRatio",
        },
     ],
     "perBucket" : True,
    },
    {"name" : "GrowthRate",
     "ingredients" : [
        {
            "name" : "dataGrowthRateForItems",
            "description" : "Data Growth rate for items",
            "counter" : "curr_items",
            "scale" : "day",
            "code" : "ItemGrowth",
            "unit" : "percentage",
        },
     ]
    },
    {"name" : "VBucketNumber",
     "ingredients" : [
        {
            "name" : "activeVbucketNumber",
            "description" : "Active VBucket number is less than expected",
            "counter" : "vb_active_num",
            "scale" : "hour",
            "code" : "NumVbuckt",
            "threshold" : 1024,
        },
        {
            "name" : "replicaVBucketNumber",
            "description" : "Replica VBucket number is less than expected",
            "counter" : "vb_replica_num",
            "scale" : "hour",
            "code" : "NumVbuckt",
            "threshold" : 1024,
        },
     ],
     "indicator" : True,
    },
    {"name" : "MemoryUsage",
     "ingredients" : [
        {
            "name" : "memoryUsage",
            "description" : "Check memory usage",
            "counter" : "mem_used",
            "scale" : "hour",
            "code" : "MemUsed",
        },
     ],
     "nodeDisparate" : True,
    },
    {"name" : "RebalancePerformance",
     "ingredients" : [
        {
            "name" : "rebalanceStuck",
            "description" : "Check if rebalance is stuck",
            "counter" : "idle",
            "code" : "RebalanceStuck",
        },
        {
            "name" : "highBackfillRemaing",
            "description" : "Tap queue backfilll remaining is too high",
            "counter" : "ep_tap_queue_backfillremaining",
            "code" : "RebalanceStuck",
            "threshold" : 1000,
        },
     ],
     "indicator" : True,
    },
    {"name" : "MemoryFragmentation",
     "ingredients" : [
        {
            "name" : "totalFragmentation",
            "description" : "Total memory fragmentation",
            "counter" : "total_fragmentation_bytes",
            "code" : "MemoryFramentation",
            "unit" : "size",
            "threshold" : 1073741824,  # 1GB
        },
        {
            "name" : "diskDelete",
            "description" : "Averge disk delete time",
            "counter" : "disk_del",
            "code" : "MemoryFramentation",
            "unit" : "time",
            "threshold" : 1000     #1ms
        },
        {
            "name" : "diskUpdate",
            "description" : "Averge disk update time",
            "counter" : "disk_update",
            "code" : "MemoryFramentation",
            "unit" : "time",
            "threshold" : 1000     #1ms
        },
        {
            "name" : "diskInsert",
            "description" : "Averge disk insert time",
            "type" : "python",
            "counter" : "disk_insert",
            "code" : "MemoryFramentation",
            "unit" : "time",
            "threshold" : 1000     #1ms
        },
        {
            "name" : "diskInsert",
            "description" : "Averge disk insert time",
            "counter" : "disk_commit",
            "code" : "MemoryFramentation",
            "unit" : "time",
            "threshold" : 5000000     #10s
        },
     ],
     "indicator" : True,
    },
    {"name" : "EPEnginePerformance",
     "ingredients" : [
        {
            "name" : "flusherState",
            "description" : "Engine flusher state",
            "counter" : "ep_flusher_state",
            "code" : "EPEnginePerformance",
            "threshold" : "running",
        },
        {
            "name" : "flusherCompleted",
            "description" : "Flusher completed",
            "counter" : "ep_flusher_num_completed",
            "code" : "EPEnginePerformance",
            "threshold" : 0
        },
        {
            "name" : "avgItemLoadTime",
            "description" : "Average item loaded time",
            "counter" : "ep_bg_load_avg",
            "code" : "EPEnginePerformance",
            "threshold" : 100,
        },
        {
            "name" : "avgItemWaitTime",
            "description" : "Averge item waited time",
            "counter" : "ep_bg_wait_avg",
            "code" : "EPEnginePerformance",
            "threshold" : 100
        },
     ],
     "indicator" : True,
    },
]



#!/usr/bin/env python

import logging
import socket
import struct
import time
import sys

import mc_bin_client
import memcacheConstants
import pump

try:
    import ctypes
except ImportError:
    cb_path = '/opt/couchbase/lib/python'
    while cb_path in sys.path:
        sys.path.remove(cb_path)
    try:
        import ctypes
    except ImportError:
        sys.exit('error: could not import ctypes module')
    else:
        sys.path.insert(0, cb_path)

OP_MAP = {
    'get': memcacheConstants.CMD_GET,
    'set': memcacheConstants.CMD_SET,
    'add': memcacheConstants.CMD_ADD,
    'delete': memcacheConstants.CMD_DELETE,
    }

OP_MAP_WITH_META = {
    'get': memcacheConstants.CMD_GET,
    'set': memcacheConstants.CMD_SET_WITH_META,
    'add': memcacheConstants.CMD_ADD_WITH_META,
    'delete': memcacheConstants.CMD_DELETE_WITH_META
    }

class MCSink(pump.Sink):
    """Dumb client sink using binary memcached protocol.
       Used when moxi or memcached is destination."""

    def __init__(self, opts, spec, source_bucket, source_node,
                 source_map, sink_map, ctl, cur):
        super(MCSink, self).__init__(opts, spec, source_bucket, source_node,
                                     source_map, sink_map, ctl, cur)

        self.op_map = OP_MAP
        if opts.extra.get("try_xwm", 1):
            self.op_map = OP_MAP_WITH_META

        self.init_worker(MCSink.run)

    def close(self):
        self.push_next_batch(None, None)

    @staticmethod
    def check_base(opts, spec):
        if getattr(opts, "destination_vbucket_state", "active") != "active":
            return ("error: only --destination-vbucket-state=active" +
                    " is supported by this destination: %s") % (spec)

        op = getattr(opts, "destination_operation", None)
        if not op in [None, 'set', 'add', 'get']:
            return ("error: --destination-operation unsupported value: %s" +
                    "; use set, add, get") % (op)

        # Skip immediate superclass Sink.check_base(),
        # since MCSink can handle different destination operations.
        return pump.EndPoint.check_base(opts, spec)

    @staticmethod
    def run(self):
        """Worker thread to asynchronously store batches into sink."""

        mconns = {} # State kept across scatter_gather() calls.

        while not self.ctl['stop']:
            batch, future = self.pull_next_batch()
            if not batch:
                self.future_done(future, 0)
                self.close_mconns(mconns)
                return

            backoff = 0.1 # Reset backoff after a good batch.

            while batch:  # Loop in case retry is required.
                rv, batch, need_backoff = self.scatter_gather(mconns, batch)
                if rv != 0:
                    self.future_done(future, rv)
                    self.close_mconns(mconns)
                    return

                if batch:
                    self.cur["tot_sink_retry_batch"] = \
                        self.cur.get("tot_sink_retry_batch", 0) + 1

                if need_backoff:
                    backoff = min(backoff * 2.0, 10.0)
                    logging.warn("backing off, secs: %s" % (backoff))
                    time.sleep(backoff)

            self.future_done(future, 0)

        self.close_mconns(mconns)

    def close_mconns(self, mconns):
        for k, conn in mconns.items():
            conn.close()

    def scatter_gather(self, mconns, batch):
        conn = mconns.get("conn")
        if not conn:
            rv, conn = self.connect()
            if rv != 0:
                return rv, None
            mconns["conn"] = conn

        # TODO: (1) MCSink - run() handle --data parameter.

        # Scatter or send phase.
        rv = self.send_msgs(conn, batch.msgs, self.operation())
        if rv != 0:
            return rv, None, None

        # Gather or recv phase.
        rv, retry, refresh = self.recv_msgs(conn, batch.msgs)
        if refresh:
            self.refresh_sink_map()
        if retry:
            return rv, batch, True

        return rv, None, None

    def send_msgs(self, conn, msgs, operation, vbucket_id=None):
        m = []

        for i, msg in enumerate(msgs):
            cmd, vbucket_id_msg, key, flg, exp, cas, meta, val = msg
            if vbucket_id is not None:
                vbucket_id_msg = vbucket_id

            if self.skip(key, vbucket_id_msg):
                continue

            rv, cmd = self.translate_cmd(cmd, operation, meta)
            if rv != 0:
                return rv

            if cmd == memcacheConstants.CMD_GET:
                val, flg, exp, cas = '', 0, 0, 0
            if cmd == memcacheConstants.CMD_NOOP:
                key, val, flg, exp, cas = '', '', 0, 0, 0

            rv, req = self.cmd_request(cmd, vbucket_id_msg, key, val,
                                       ctypes.c_uint32(flg).value,
                                       exp, cas, meta, i)
            if rv != 0:
                return rv

            self.append_req(m, req)

        if m:
            try:
                conn.s.send(''.join(m))
            except socket.error, e:
                return "error: conn.send() exception: %s" % (e)

        return 0

    def recv_msgs(self, conn, msgs, vbucket_id=None):
        refresh = False
        retry = False

        for i, msg in enumerate(msgs):
            cmd, vbucket_id_msg, key, flg, exp, cas, meta, val = msg
            if vbucket_id is not None:
                vbucket_id_msg = vbucket_id

            if self.skip(key, vbucket_id_msg):
                continue

            try:
                r_cmd, r_status, r_ext, r_key, r_val, r_cas, r_opaque = \
                    self.read_conn(conn)

                if i != r_opaque:
                    return "error: opaque mismatch: %s %s" % (i, r_opaque), None, None

                if r_status == memcacheConstants.ERR_SUCCESS:
                    continue
                elif r_status == memcacheConstants.ERR_KEY_EEXISTS:
                    logging.warn("item exists: %s, key: %s" %
                                 (self.spec, key))
                    continue
                elif r_status == memcacheConstants.ERR_KEY_ENOENT:
                    if (cmd != memcacheConstants.CMD_TAP_DELETE and
                        cmd != memcacheConstants.CMD_GET):
                        logging.warn("item not found: %s, key: %s" %
                                     (self.spec, key))
                    continue
                elif (r_status == memcacheConstants.ERR_ETMPFAIL or
                      r_status == memcacheConstants.ERR_EBUSY or
                      r_status == memcacheConstants.ERR_ENOMEM):
                    retry = True # Retry the whole batch again next time.
                    continue     # But, finish recv'ing current batch.
                elif r_status == memcacheConstants.ERR_NOT_MY_VBUCKET:
                    msg = ("received NOT_MY_VBUCKET;"
                           " perhaps the cluster is/was rebalancing;"
                           " vbucket_id: %s, key: %s, spec: %s, host:port: %s:%s"
                           % (vbucket_id_msg, key, self.spec,
                              conn.host, conn.port))
                    if self.opts.extra.get("nmv_retry", 1):
                        logging.warn("warning: " + msg)
                        refresh = True
                        retry = True
                        self.cur["tot_sink_not_my_vbucket"] = \
                            self.cur.get("tot_sink_not_my_vbucket", 0) + 1
                    else:
                        return "error: " + msg, None, None
                elif r_status == memcacheConstants.ERR_UNKNOWN_COMMAND:
                    if self.op_map == OP_MAP:
                        if not retry:
                            return "error: unknown command: %s" % (r_cmd), None, None
                    else:
                        if not retry:
                            logging.warn("destination does not take XXX-WITH-META"
                                         " commands; will use META-less commands")
                        self.op_map = OP_MAP
                        retry = True
                else:
                    return "error: MCSink MC error: " + str(r_status), None, None

            except Exception, e:
                logging.error("MCSink exception: %s", e)
                return "error: MCSink exception: " + str(e), None, None

        return 0, retry, refresh

    def translate_cmd(self, cmd, op, meta):
        if len(str(meta)) <= 0:
            # The source gave no meta, so use regular commands.
            self.op_map = OP_MAP

        if cmd == memcacheConstants.CMD_TAP_MUTATION:
            m = self.op_map.get(op, None)
            if m:
                return 0, m
            return "error: MCSink.translate_cmd, unsupported op: " + op, None

        if cmd == memcacheConstants.CMD_TAP_DELETE:
            if op == 'get':
                return 0, memcacheConstants.CMD_NOOP
            return 0, self.op_map['delete']

        if cmd == memcacheConstants.CMD_GET:
            return 0, cmd

        return "error: MCSink - unknown cmd: %s, op: %s" % (cmd, op), None

    def append_req(self, m, req):
        hdr, ext, key, val = req
        m.append(hdr)
        if ext:
            m.append(ext)
        if key:
            m.append(str(key))
        if val:
            m.append(str(val))

    @staticmethod
    def can_handle(opts, spec):
        return (spec.startswith("memcached://") or
                spec.startswith("memcached-binary://"))

    @staticmethod
    def check(opts, spec, source_map):
        host, port, user, pswd, path = \
            pump.parse_spec(opts, spec, int(getattr(opts, "port", 11211)))
        rv, conn = MCSink.connect_mc(host, port, user, pswd)
        if rv != 0:
            return rv, None
        conn.close()
        return 0, None

    def refresh_sink_map(self):
        return 0

    @staticmethod
    def consume_design(opts, sink_spec, sink_map,
                       source_bucket, source_map, source_design):
        if source_design:
            logging.warn("warning: cannot restore bucket design"
                         " on a memached destination")
        return 0

    def consume_batch_async(self, batch):
        return self.push_next_batch(batch, pump.SinkBatchFuture(self, batch))

    def connect(self):
        host, port, user, pswd, path = \
            pump.parse_spec(self.opts, self.spec,
                            int(getattr(self.opts, "port", 11211)))
        return MCSink.connect_mc(host, port, user, pswd)

    @staticmethod
    def connect_mc(host, port, user, pswd):
        mc = mc_bin_client.MemcachedClient(host, int(port))
        if user:
            try:
                mc.sasl_auth_plain(str(user), str(pswd))
            except EOFError:
                return "error: SASL auth error: %s:%s, user: %s" % \
                    (host, port, user), None
            except mc_bin_client.MemcachedError:
                return "error: SASL auth failed: %s:%s, user: %s" % \
                    (host, port, user), None
            except socket.error:
                return "error: SASL auth exception: %s:%s, user: %s" % \
                    (host, port, user), None
        return 0, mc

    def cmd_request(self, cmd, vbucket_id, key, val, flg, exp, cas, meta, opaque):
        if (cmd == memcacheConstants.CMD_SET_WITH_META or
            cmd == memcacheConstants.CMD_ADD_WITH_META or
            cmd == memcacheConstants.CMD_DELETE_WITH_META):
            if meta:
                seq_no = str(meta)
                if len(seq_no) > 8:
                    seq_no = seq_no[0:8]
                if len(seq_no) < 8:
                    # The seq_no might be 32-bits from 2.0DP4, so pad with 0x00's.
                    seq_no = ('\x00\x00\x00\x00\x00\x00\x00\x00' + seq_no)[-8:]
                ext = (struct.pack(">II", flg, exp) + seq_no +
                       struct.pack(">Q", cas))
            else:
                ext = struct.pack(">IIQQ", flg, exp, 0, cas)
        elif (cmd == memcacheConstants.CMD_SET or
              cmd == memcacheConstants.CMD_ADD):
            ext = struct.pack(memcacheConstants.SET_PKT_FMT, flg, exp)
        elif (cmd == memcacheConstants.CMD_DELETE or
              cmd == memcacheConstants.CMD_GET or
              cmd == memcacheConstants.CMD_NOOP):
            ext = ''
        else:
            return "error: MCSink - unknown cmd for request: " + str(cmd), None

        hdr = self.cmd_header(cmd, vbucket_id, key, val, ext, 0, opaque)
        return 0, (hdr, ext, key, val)

    def cmd_header(self, cmd, vbucket_id, key, val, ext, cas, opaque,
                   dtype=0,
                   fmt=memcacheConstants.REQ_PKT_FMT,
                   magic=memcacheConstants.REQ_MAGIC_BYTE):
        return struct.pack(fmt, magic, cmd,
                           len(key), len(ext), dtype, vbucket_id,
                           len(key) + len(ext) + len(val), opaque, cas)

    def read_conn(self, conn):
        ext = ''
        key = ''
        val = ''

        buf, cmd, errcode, extlen, keylen, data, cas, opaque = \
            self.recv_msg(conn.s, getattr(conn, 'buf', ''))
        conn.buf = buf

        if data:
            ext = data[0:extlen]
            key = data[extlen:extlen+keylen]
            val = data[extlen+keylen:]

        return cmd, errcode, ext, key, val, cas, opaque

    def recv_msg(self, sock, buf):
        pkt, buf = self.recv(sock, memcacheConstants.MIN_RECV_PACKET, buf)
        if not pkt:
            raise EOFError()
        magic, cmd, keylen, extlen, dtype, errcode, datalen, opaque, cas = \
            struct.unpack(memcacheConstants.RES_PKT_FMT, pkt)
        if magic != memcacheConstants.RES_MAGIC_BYTE:
            raise Exception("unexpected recv_msg magic: " + str(magic))
        data, buf = self.recv(sock, datalen, buf)
        return buf, cmd, errcode, extlen, keylen, data, cas, opaque

    def recv(self, skt, nbytes, buf):
        while len(buf) < nbytes:
            data = None
            try:
                data = skt.recv(max(nbytes - len(buf), 4096))
            except socket.timeout:
                logging.error("error: recv socket.timeout")
            except Exception, e:
                logging.error("error: recv exception: " + str(e))

            if not data:
                return None, ''
            buf += data

        return buf[:nbytes], buf[nbytes:]



#!/usr/bin/python

try: import simplejson as json
except ImportError: import json
import couchdb
import httplib
import urllib
import common
import unittest
import time


HOST = "localhost:5984"
SET_NAME = "test_suite_set_view"
NUM_PARTS = 8
NUM_DOCS = 400000
DDOC = {
    "_id": "_design/test",
    "language": "javascript",
    "views": {
        "mapview": {
            "map": "function(doc) { emit(doc.integer, doc.string); }"
        },
        "redview": {
            "map": "function(doc) { emit(doc.integer, doc.string); }",
            "reduce": "_count"
        }
    }
}


class TestReplicaIndex(unittest.TestCase):

    def setUp(self):
        self._params = {
            "host": HOST,
            "ddoc": DDOC,
            "nparts": NUM_PARTS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME,
            "server": couchdb.Server(url = "http://" + HOST)
            }
        # print "Creating databases"
        common.create_dbs(self._params)
        common.populate(self._params)
        # print "Databases created"
        # print "Configuring set view with:"
        # print "\tmaximum of 8 partitions"
        # print "\tactive partitions = [0, 1, 2, 3]"
        # print "\tpassive partitions = []"
        # print "\treplica index = True"
        active_parts = [0, 1, 2, 3]
        common.define_set_view(self._params, active_parts, [], True)


    def tearDown(self):
        # print "Deleting test data"
        common.create_dbs(self._params, True)


    def test_replica_index(self):
        # print "Querying map view in steady state"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])

        self.assertEqual(view_result["total_rows"], expected_row_count,
                         "Query returned %d total_rows" % expected_row_count)
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        # print "verifying group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [],
                         "main group has [] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [],
                         "main group has [] as replicas on transfer")
        self.assertEqual(sorted(info["update_seqs"].keys()), ["0", "1", "2", "3"],
                         "right keys in update_seqs of the main index")
        for i in [0, 1, 2, 3]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d" % (expected_seq, i + 1))
        self.assertEqual(type(info["replica_group_info"]), dict, "group has replica_group_info")
        self.assertEqual(info["replica_group_info"]["max_number_partitions"], 8,
                          "replica group defined with maximum of 8 partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [],
                          "replica group has no passive partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")
        self.assertEqual(info["replica_group_info"]["update_seqs"].keys(), [],
                          "replica group update_seqs is empty")
        self.assertEqual(info["replica_group_info"]["stats"]["partial_updates"], 0,
                          "replica group has 0 partial updates")
        self.assertEqual(info["replica_group_info"]["stats"]["full_updates"], 0,
                          "replica group has 0 full updates")

        # print "defining partitions [4, 5, 6, 7] as replicas"
        common.add_replica_partitions(self._params, [4, 5, 6, 7])

        # print "verifying group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [],
                         "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [4, 5, 6, 7],
                          "replica group has [4, 5, 6, 7] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "waiting for replica index to update"
        seconds = 0
        while info["replica_group_info"]["stats"]["full_updates"] < 1:
            time.sleep(5)
            seconds += 5
            if seconds > 900:
                raise(Exception("timeout waiting for replica index full update"))
            info = common.get_set_view_info(self._params)

        self.assertEqual(sorted(info["replica_group_info"]["update_seqs"].keys()), ["4", "5", "6", "7"],
                          "replica group update_seqs has keys [4, 5, 6, 7]")
        for i in [4, 5, 6, 7]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["replica_group_info"]["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for replica partition %d" % (expected_seq, i + 1))

        # print "Querying map view after marking partitions [4, 5, 6, 7] as replicas (not yet active)"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Querying reduce view, with ?group=false, after marking partitions [4, 5, 6, 7] as replicas (not yet active)"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview")

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after marking partitions [4, 5, 6, 7] as replicas (not yet active)"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        new_info = common.get_set_view_info(self._params)
        common.restart_server(self._params)

        # print "Querying map view again after immediately after restarting the server"
        (resp, view_result_after) = common.query(self._params, "mapview")
        self.assertEqual(view_result_after["rows"], view_result["rows"],
                         "Same view rows after server restart")

        # print "Querying reduce view, with ?group=false, after restarting the server"
        (resp, red_view_result_nogroup_after) = common.query(self._params, "redview")
        self.assertEqual(red_view_result_nogroup_after["rows"],
                         red_view_result_nogroup["rows"],
                         "Same view result as before")

        # print "Querying reduce view, with ?group=true, after restarting the server"
        (resp, red_view_result_group_after) = common.query(self._params, "redview", {"group": "true"})
        self.assertEqual(red_view_result_group_after["rows"],
                         red_view_result_group["rows"],
                         "Same view result as before")

        # print "removing partitions [5, 7] from replica set"
        common.remove_replica_partitions(self._params, [5, 7])

        # print "verifying group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [4, 6],
                         "main group has [4, 6] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [],
                         "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [4, 6],
                          "replica group has [4, 6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [5, 7],
                          "replica group has [5, 7] as cleanup partitions")

        # print "waiting for replica index to finish cleanup"
        seconds = 0
        while info["replica_group_info"]["stats"]["cleanups"] < 1:
            time.sleep(5)
            seconds += 5
            if seconds > 900:
                raise(Exception("timeout waiting for replica index cleanup"))
            info = common.get_set_view_info(self._params)

        self.assertEqual(info["replica_group_info"]["passive_partitions"], [4, 6],
                          "replica group has [4, 6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has [] as cleanup partitions")
        self.assertEqual(sorted(info["replica_group_info"]["update_seqs"].keys()), ["4", "6"],
                          "replica group update_seqs has keys [4, 6")

        new_info = common.get_set_view_info(self._params)
        self.restart_compare_info(info, new_info)

        info = common.get_set_view_info(self._params)
        updates_before = info["replica_group_info"]["stats"]["full_updates"]

        # print "defining partitions [5, 7] as replicas again"
        common.add_replica_partitions(self._params, [5, 7])

        # replica group likely has [5, 7] in a pending transition, just wait a while if so
        self.wait_for_pending_transition_applied()

        # print "verifying group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [],
                         "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [4, 5, 6, 7],
                          "replica group has [4, 5, 6, 7] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "waiting for replica index to index partitions 5 and 7"
        seconds = 0
        while info["replica_group_info"]["stats"]["full_updates"] < (updates_before + 1):
            time.sleep(5)
            seconds += 5
            if seconds > 900:
                raise(Exception("timeout waiting for replica index full update"))
            info = common.get_set_view_info(self._params)

        self.assertEqual(sorted(info["replica_group_info"]["update_seqs"].keys()), ["4", "5", "6", "7"],
                          "replica group update_seqs has keys [4, 5, 6, 7]")
        for i in [4, 5, 6, 7]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["replica_group_info"]["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for replica partition %d" % (expected_seq, i + 1))

        # print "Querying map view after marking partitions [5, 7] as replicas again (not yet active)"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Querying reduce view, with ?group=false, after marking partitions [5, 7] as replicas again (not yet active)"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview")

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after marking partitions [5, 7] as replicas again (not yet active)"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Marking partitions [4, 5, 7] as active (to be transferred from replica to main index)"
        common.set_partition_states(self._params, active = [4, 5, 7])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [4, 5, 7], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [4, 5, 7],
                         "main group has [4, 5, 7] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [4, 5, 7],
                          "replica group has [4, 5, 7] as active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "Querying map view with ?stale=update_after to trigger transferral of" + \
        #     " replica partitions [4, 5, 7] from replica index to main index"

        updates_before = info["stats"]["full_updates"]

        (resp, view_result) = common.query(self._params, "mapview", {"stale": "update_after"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 7])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(7, common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7]), self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")

        # print "Querying reduce view, with ?group=false, after triggering transfer of replica partitions [4, 5, 7]"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 7])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after triggering transfer of replica partitions [4, 5, 7]"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 7])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")

        # print "Waiting for transferral of replica partitions [4, 5, 7] into main index"
        info = common.get_set_view_info(self._params)
        seconds = 0
        while info["stats"]["full_updates"] < (updates_before + 1):
            time.sleep(5)
            seconds += 5
            if seconds > 900:
                raise(Exception("timeout waiting for transfer of replica partitions [4, 5, 7]"))
            info = common.get_set_view_info(self._params)

        self.assertEqual(info["active_partitions"], [0, 1, 2, 3, 4, 5, 7], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [6], "main group has [6] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [], "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has [] as active partitions")
        self.assertEqual(sorted(info["update_seqs"].keys()), ["0", "1", "2", "3", "4", "5", "7"],
                         "Right list of update seqs for main group")
        self.assertEqual(sorted(info["replica_group_info"]["update_seqs"].keys()), ["6"],
                         "Right list of update seqs for replica group")
        for i in [0, 1, 2, 3, 4, 5, 7]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d in main group" % (expected_seq, i + 1))
        for i in [6]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["replica_group_info"]["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d in replica group" % (expected_seq, i + 1))

        # print "Querying map view after replica partitions were transferred"
        (resp, view_result_after) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 7])
        self.assertEqual(len(view_result_after["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        self.assertEqual(len(view_result_after["rows"]), len(view_result["rows"]),
                         "Same number of view rows after replica partitions were transferred")
        for i in xrange(len(view_result_after["rows"])):
            self.assertEqual(view_result_after["rows"][i], view_result["rows"][i],
                             "Row %d has same content after replica partitions were transferred" % (i + 1))

        # print "Querying reduce view, with ?group=false, after replica partitions were transferred"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 7])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after replica partitions were transferred"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 7])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")

        # print "Marking partitions [4, 5, 7] for cleanup in the main group"
        common.set_partition_states(self._params, cleanup = [4, 5, 7])

        # print "Waiting for cleanup (compaction) to finish"
        common.compact_set_view(self._params)

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [6], "main group has [6] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [], "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has [] active partitions")

        # print "Querying map view after marking partitions [4, 5, 7] for cleanup in the main group"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Querying reduce view, with ?group=false, after marking partitions [4, 5, 7] for cleanup in the main group"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after marking partitions [4, 5, 7] for cleanup in the main group"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "defining partitions [4, 5, 7] as replicas again"
        common.add_replica_partitions(self._params, [4, 5, 7])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [],
                         "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [4, 5, 6, 7],
                          "replica group has [4, 5, 6, 7] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "removing partitions [4, 5, 7] from replica set again"
        common.remove_replica_partitions(self._params, [4, 5, 7])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [6], "main group has [6] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [], "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [4, 5, 7],
                          "replica group has [4, 5, 7] as cleanup partitions")

        # print "defining partitions [4, 5, 7] as replicas again"
        common.add_replica_partitions(self._params, [4, 5, 7])
        self.wait_for_pending_transition_applied()

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [],
                         "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [4, 5, 6, 7],
                          "replica group has [4, 5, 6, 7] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "Querying map view after marking partitions [4, 5, 7] as replicas (not active)"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Querying reduce view, with ?group=false, after marking partitions [4, 5, 7] as replicas (not active)"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after marking partitions [4, 5, 7] as replicas (not active)"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Marking partitions [4, 5, 7] as active (to be transferred from replica to main index)"
        common.set_partition_states(self._params, active = [4, 5, 7])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [4, 5, 7], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [4, 5, 7],
                         "main group has [4, 5, 7] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [4, 5, 7],
                          "replica group has [4, 5, 7] as active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "Querying view with ?stale=update_after to trigger transferral of" + \
        #    " replica partitions [4, 5, 7] from replica index to main index"

        updates_before = info["stats"]["full_updates"]

        (resp, view_result) = common.query(self._params, "mapview", {"stale": "update_after", "limit": "1"})

        # print "unmarking partitions [4, 5] as replicas while partitions [4, 5, 7] are being transferred from replica to main group"
        common.remove_replica_partitions(self._params, [4, 5])

        # print "Waiting for transferral of replica partitions [7] into main index"
        info = common.get_set_view_info(self._params)
        seconds = 0
        while info["stats"]["full_updates"] < (updates_before + 1):
            time.sleep(5)
            seconds += 5
            if seconds > 900:
                raise(Exception("timeout waiting for transfer of replica partitions [7]"))
            info = common.get_set_view_info(self._params)

        self.assertEqual(info["active_partitions"], [0, 1, 2, 3, 7], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [6], "main group has [6] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [], "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has [] as active partitions")
        self.assertEqual(sorted(info["update_seqs"].keys()), ["0", "1", "2", "3", "7"],
                         "Right list of update seqs for main group")
        self.assertEqual(sorted(info["replica_group_info"]["update_seqs"].keys()), ["6"],
                         "Right list of update seqs for replica group")
        for i in [0, 1, 2, 3, 7]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d in main group" % (expected_seq, i + 1))
        for i in [6]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["replica_group_info"]["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d in replica group" % (expected_seq, i + 1))

        # print "Querying map view after removing replica partitions [4, 5]"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 7])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 7 in view result")

        # print "Querying reduce view, with ?group=false, after removing replica partitions [4, 5]"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3, 7])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after removing replica partitions [4, 5]"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 7])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 7 in view result")

        # print "Marking partitions [7] for cleanup in the main group"
        common.set_partition_states(self._params, cleanup = [7])

        # print "Waiting for cleanup (compaction) to finish"
        common.compact_set_view(self._params)

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [6], "main group has [6] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [], "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has [] active partitions")

        # print "Querying map view after removing replica partitions [7]"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Querying reduce view, with ?group=false, after removing replica partitions [7]"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after removing replica partitions [7]"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "defining partitions [4, 5, 7] as replicas again"
        common.add_replica_partitions(self._params, [4, 5, 7])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [],
                         "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [4, 5, 6, 7],
                          "replica group has [4, 5, 6, 7] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has no active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "Querying map view after marking partitions [4, 5, 7] as replicas (not yet active)"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Querying reduce view, with ?group=false, after marking partitions [4, 5, 7] as replicas (not yet active)"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after marking partitions [4, 5, 7] as replicas (not yet active)"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 7 not in view result")

        # print "Marking partitions [4, 5, 7] as active (to be transferred from replica to main index)"
        common.set_partition_states(self._params, active = [4, 5, 7])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [4, 5, 7], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [4, 5, 6, 7],
                         "main group has [4, 5, 6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [4, 5, 7],
                         "main group has [4, 5, 7] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [4, 5, 7],
                          "replica group has [4, 5, 7] as active partitions")
        self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [],
                          "replica group has no cleanup partitions")

        # print "Querying view with ?stale=update_after to trigger transferral of" + \
        #     " replica partitions [4, 5, 7] from replica index to main index"

        updates_before = info["stats"]["full_updates"]

        (resp, view_result) = common.query(self._params, "mapview", {"stale": "update_after", "limit": "1"})

        # print "marking partitions [4, 5] for cleanup while partitions [4, 5, 7] are being transferred from replica to main group"
        common.set_partition_states(self._params, cleanup = [4, 5])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [7], "right passive partitions list")
        # Not tested due to race conditions
        # self.assertEqual(info["cleanup_partitions"], [4, 5], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [6, 7],
                         "main group has [6, 7] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [7],
                         "main group has [7] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [7],
                          "replica group has [7] as active partitions")
        # Not tested due to race conditions
        # self.assertEqual(info["replica_group_info"]["cleanup_partitions"], [4, 5],
        #                  "replica group has [4, 5] as cleanup partitions")

        # print "Waiting for transferral of replica partitions [7] into main index"
        info = common.get_set_view_info(self._params)
        seconds = 0
        while info["stats"]["full_updates"] < (updates_before + 1):
            time.sleep(5)
            seconds += 5
            if seconds > 900:
                raise(Exception("timeout waiting for transfer of replica partitions [7]"))
            info = common.get_set_view_info(self._params)

        self.assertEqual(info["active_partitions"], [0, 1, 2, 3, 7], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["replica_partitions"], [6], "main group has [6] as replica partitions")
        self.assertEqual(info["replicas_on_transfer"], [], "main group has [] as replicas on transfer")
        self.assertEqual(info["replica_group_info"]["passive_partitions"], [6],
                          "replica group has [6] as passive partitions")
        self.assertEqual(info["replica_group_info"]["active_partitions"], [],
                          "replica group has [] as active partitions")
        self.assertEqual(sorted(info["update_seqs"].keys()), ["0", "1", "2", "3", "7"],
                         "Right list of update seqs for main group")
        self.assertEqual(sorted(info["replica_group_info"]["update_seqs"].keys()), ["6"],
                         "Right list of update seqs for replica group")
        for i in [0, 1, 2, 3, 7]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d in main group" % (expected_seq, i + 1))
        for i in [6]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["replica_group_info"]["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d in replica group" % (expected_seq, i + 1))

        # print "Querying map view after marking partitions [4, 5] for cleanup"
        (resp, view_result) = common.query(self._params, "mapview")

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 7])
        self.assertEqual(len(view_result["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(view_result)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in view_result["rows"]:
            all_keys[row["key"]] = row["value"]
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 7 in view result")

        # print "Querying reduce view, with ?group=false, after marking partitions [4, 5] for cleanup"
        (resp, red_view_result_nogroup) = common.query(self._params, "redview", {"stale": "ok"})

        expected_reduce_value = common.set_doc_count(self._params, [0, 1, 2, 3, 7])
        self.assertEqual(len(red_view_result_nogroup["rows"]), 1, "Query returned 1 row")
        self.assertTrue(red_view_result_nogroup["rows"][0]["key"] is None, "Key is null")
        self.assertEqual(red_view_result_nogroup["rows"][0]["value"], expected_reduce_value,
                         "Reduce value is %d with ?group=false" % expected_reduce_value)

        # print "Querying reduce view, with ?group=true, after marking partitions [4, 5] for cleanup"
        (resp, red_view_result_group) = common.query(self._params, "redview", {"group": "true", "stale": "ok"})

        expected_row_count = common.set_doc_count(self._params, [0, 1, 2, 3, 7])
        self.assertEqual(len(red_view_result_group["rows"]), expected_row_count,
                         "Query returned %d rows" % expected_row_count)
        common.test_keys_sorted(red_view_result_group)

        doc_count = common.set_doc_count(self._params, [0, 1, 2, 3, 4, 5, 6, 7])
        all_keys = {}
        for row in red_view_result_group["rows"]:
            all_keys[row["key"]] = row["value"]
            self.assertEqual(row["value"], 1, "Value for key %d is 1" % row["key"])
        for key in xrange(1, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 0 in view result")
        for key in xrange(2, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 1 in view result")
        for key in xrange(3, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 2 in view result")
        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 3 in view result")
        for key in xrange(5, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 4 not in view result")
        for key in xrange(6, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 5 not in view result")
        for key in xrange(7, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key from partition 6 not in view result")
        for key in xrange(8, doc_count, self._params["nparts"]):
            self.assertTrue(key in all_keys, "Key from partition 7 in view result")


    def restart_compare_info(self, info_before, info_after):
        # print "Restarting server"
        time.sleep(1)
        common.restart_server(self._params)

        del info_before["stats"]
        del info_before["replica_group_info"]["stats"]
        del info_after["stats"]
        del info_after["replica_group_info"]["stats"]
        self.assertEqual(info_after, info_before, "same index state after server restart")


    def wait_for_pending_transition_applied(self):
        # print "Waiting for pending transition to be applied"
        iterations = 0
        while True:
            if iterations > 600:
                raise(Exception("timeout waiting for pending transition to be applied"))
            info = common.get_set_view_info(self._params)
            if (info["pending_transition"] is None) and \
                (info["replica_group_info"]["pending_transition"] is None):
                break
            else:
                time.sleep(1)
                iterations += 1
